'use strict'

const moment = require("moment");
const EventEmitter = require('events');
const fastChunkString = require('@shelf/fast-chunk-string');
const Util = require("../util/util");
const Connection = require("../model/connection");
const ConnectionEvent = require("../model/connectionevent");
const ConnectionStatus = require("../model/connectionstatus");
const ConnectorStatus = require("../model/connectorstatus");
const ConnectionData = require("../model/connectiondata");
const PingBody = require("../model/connectiondatabody/pingbody");
const ConnectBody = require("../model/connectiondatabody/connectbody");
const ErrorBody = require("../model/connectiondatabody/errorbody");
const AnswerBody = require("../model/connectiondatabody/answerbody");
const RequestBody = require("../model/connectiondatabody/requestbody");
const ResponseBody = require("../model/connectiondatabody/responsebody");
const OfferBody = require("../model/connectiondatabody/offerbody");
const ChunkDataBody = require("../model/connectiondatabody/chunkdatabody");
const ChunkEndBody = require("../model/connectiondatabody/chunkendbody");
const ChunkStartBody = require("../model/connectiondatabody/chunkstartbody");
const Chunk = require("../model/chunk");
const Encoder = require("../util/encoder");
const DisconnectBody = require("../model/connectiondatabody/disconnectbody");

module.exports = class Connector {

    static _instance;
    #signalserver;
    #genesisAnswerCollectInterval;
    #collectingGenesisAnswer;
    #pingingPeer;
    initiator;
    eventEmitter;
    wrtc;
    genesisConnectionId;
    genesisConnectionOffer;
    connections = {};
    lastPing;
    thread = 0;
    threadCount = 3;
    readyConnectionCount = 3;
    openConnectionCount = 0;
    maxDataSize = 200 * 1024;
    status = ConnectorStatus.initiated;
    data = {
        send: [],
        failed: [],
    }
    count = {
        sent: 0,
    }

    constructor({signalserver, initiator= true, wrtc = null}) {
        try {
            if(Connector._instance) return Connector._instance;
            Connector._instance = this;

            this.#signalserver = signalserver;
            this.wrtc = wrtc;
            this.initiator = initiator;

            if(typeof initiator !== 'boolean') throw 'Invalid initiator type';
            if(Util.IsBrowser()) this.eventEmitter = new EventTarget();
            else this.eventEmitter = new EventEmitter();

            this.On('genesis_connection', async (event) => {
                try {
                    this.genesisConnectionOffer = await event.detail.connection.GetOffer();
                    this.#Emit('genesis_offer', this.genesisConnectionOffer);
                } catch (e) {
                    this.#Emit('error', {e});
                }
            });
            this.On('genesis_offer_register', (event) => {
                try {
                    this.#CollectGenesisAnswer(event.detail.waitSeconds);
                } catch (e) {
                    throw e;
                }
            });
            this.On('genesis_answer_received', async (event) => {
                try {
                    const data = event.detail;
                    const connection = this.connections[data.connectionId];
                    await connection.SetAnswer(data.answer).then(res => {
                        if(!res || res.answer.length === 0) {
                            throw {
                                msg: 'Failed to set answer',
                                connection: res
                            };
                        }
                        this.#Emit('genesis_answer_set', { connection: res });
                    }).catch(e => {
                        throw e;
                    });
                } catch (e) {
                    this.#Emit('error', {e});
                }
            });
            this.On('disconnect', (event) => {
                try {
                    this.Disconnect();
                } catch (e) {
                    this.#Emit('error', {e});
                }
            });
            this.On('ready', (event) => {
                try {
                    if(!this.#pingingPeer) {
                        this.#pingingPeer = setInterval(async () => {
                            try {
                                if(this.lastPing + 10000 < moment.now()) throw { code: 'ERR_LOSTPING' };

                                const connection = await this.#GetAliveConnection();
                                connection.Send(new ConnectionData({
                                    method: 'ping',
                                    data: new PingBody({
                                        connectionId: connection.id,
                                        peerId: connection.peerId
                                    })
                                }));
                            } catch (e) {
                                clearInterval(this.#pingingPeer);
                                this.#pingingPeer = null;
                                if(this.status !== ConnectorStatus.disconnected && this.status !== ConnectorStatus.closing) {
                                    this.#Emit('disconnect', {e});
                                }
                            }
                        }, 5000);
                    }
                } catch (e) {
                    this.#Emit('error', {e});
                }
            });

            this.#InitConnection({initiator}, initiator);

        } catch (e) {
            throw e;
        }
    }

    /**
     * register genesis connection offer to signal server
     * @param offer
     * @param waitSeconds seconds to wait for genesis connection answer
     * @constructor
     */
    async RegisterGenesisOffer({connectionOffer = null, waitSeconds = 60}) {
        try {
            if(!!connectionOffer) this.genesisConnectionOffer = connectionOffer;
            if(this.status === ConnectorStatus.disconnected) {
                await this.Reset().then(res => {
                    this.status = ConnectorStatus.initiated;
                });
            }
            if(!this.genesisConnectionOffer) throw 'genesis connection offer not set.';

            await this.CancelGenesisOffer();

            this.#signalserver.SetConnection(this.genesisConnectionOffer).then(res => {
                if(!res) {
                    throw {
                        msg: 'Failed to register genesis offer',
                        code: 'ERR_REGISTEROFFER',
                        data,
                        res
                    };
                }
                this.#Emit('genesis_offer_register', {
                    connectionId: this.genesisConnectionId,
                    offer: this.genesisConnectionOffer,
                    waitSeconds
                });
            }).catch(e => {
                throw e;
            });
        } catch (e) {
            this.#Emit('error', {e});
        }
    }

    /**
     * cancel genesis connection offer
     * @constructor
     */
    async CancelGenesisOffer() {
        try {
            if(this.#collectingGenesisAnswer === false) return;
            if(this.#genesisAnswerCollectInterval) clearInterval(this.#genesisAnswerCollectInterval);

            this.#collectingGenesisAnswer = false;
            return await this.#signalserver.SetConnection({
                connectionId: this.genesisConnectionId,
                offer: null,
                answer: null
            });
        } catch (e) {
            this.#Emit('error', {e});
        }
    }

    /**
     * send request to peer
     * @param type
     * @param body
     * @param encode
     * @param encrypt
     * @param password
     * @returns {Promise<void>}
     * @constructor
     */
    async Request(type, body, encode = true, encrypt = false, password = null) {
        try {
            const connection = await this.#GetAliveConnection();
            await connection.Send(new ConnectionData({
                method: 'request',
                data: new RequestBody({type, body}),
                encode: encode,
                encrypt: encrypt,
                password: password
            }));
        } catch (e) {
            this.#Emit('error', {e});
        }
    }

    /**
     * send response to peer request
     * @param type
     * @param body
     * @param encode
     * @param encrypt
     * @param password
     * @returns {Promise<void>}
     * @constructor
     */
    async Response(type, body, encode = true, encrypt = false, password = null) {
        try {
            const connection = await this.#GetAliveConnection();
            await connection.Send(new ConnectionData({
                method: 'response',
                data: new ResponseBody({type, body}),
                encode: encode,
                encrypt: encrypt,
                password: password
            }));
        } catch (e) {
            this.#Emit('error', {e});
        }
    }

    /**
     * send prepared array of tx objects
     * @param type
     * @param dataSequence: array of tx objects
     * @returns {Promise<void>}
     * @constructor
     */
    async SendDataSequence(type = 'send_tx', dataSequence = null) {
        let interval;
        try {
            if(this.status === 'initiated') {
                throw {
                    msg: 'Connector is not ready',
                    code: 'ERR_IMMATURE'
                }
            }
            if(this.status === ConnectorStatus.sending) {
                throw {
                    msg: 'Already sending data',
                    code: 'ERR_SENDING'
                };
            }
            if(!dataSequence && this.data.send.length === 0) {
                throw {
                    msg: 'Nothing to send',
                    code: 'ERR_EMPTY'
                };
            }
            if(Array.isArray(dataSequence) && dataSequence.length > 0 && this.data.length > 0) {
                throw {
                    msg: 'Connector is busy.',
                    code: 'ERR_BUSY'
                }
            }
            this.status = ConnectorStatus.sending;
            if(this.data.send.length === 0) {
                await this.#SetDataSequence(dataSequence).then(async sequence => {
                    const startData = new ConnectionData({
                        method: 'chunk',
                        data: new ChunkStartBody({
                            type: type,
                            length: sequence.length,
                            timestamp: moment.now(),
                        })
                    });
                    const connection = await this.#GetAliveConnection();
                    await connection.Send(startData).then(res => {
                        this.#Emit('chunk_start', startData.data);
                    });
                });
            }

            interval = setInterval(() => {
                if(this.data.send.length === 0) {
                    this.thread = 0;
                    this.status = ConnectorStatus.ready;
                    clearInterval(interval);
                    return;
                }
                if(this.data.failed.length > 0) this.thread -= this.data.failed.length;
                if(this.thread >= this.threadCount) return;
                for (const key in this.connections) {
                    if(this.connections[key].status === ConnectionStatus.disconnected) delete this.connections[key];
                    if(this.thread < this.threadCount && this.connections[key].status === ConnectionStatus.connected && this.connections[key].isSending === false) {
                        this.#SendData(this.connections[key], type);
                    }
                }
            }, 200);
        } catch (e) {
            if(interval) clearInterval(interval);
            this.status = ConnectorStatus.ready;
            // this.#GetAliveConnectionCount(true).then(count => {
            //     if(count < this.threadCount) this.#InitConnection({initiator: true}, false);
            // });
            if(this.openConnectionCount < this.threadCount) this.#InitConnection({initiator: true}, false);
            if(e.code === 'ERR_SENDING' || this.data.failed.length > 0) {
                await Util.Sleep(1000);
                if(this.data.send.length > 0) this.SendDataSequence();
            } else {
                this.#Emit('error', {e});
            }
        }
    }

    /**
     * send prepared chunks of tx object array
     * @param type
     * @param dataChunkSequence
     * @returns {Promise<void>}
     * @constructor
     */
    async SendDataChunkSequence(type = 'send_tx', dataChunkSequence = null) {
        let interval;
        try {
            if(this.status === 'initiated') {
                throw {
                    msg: 'Connector is not ready',
                    code: 'ERR_IMMATURE'
                }
            }
            if(this.status === ConnectorStatus.sending) {
                throw {
                    msg: 'Already sending data',
                    code: 'ERR_SENDING'
                };
            }
            if(!dataChunkSequence && this.data.send.length === 0) {
                throw {
                    msg: 'Nothing to send',
                    code: 'ERR_EMPTY'
                };
            }
            if(Array.isArray(dataChunkSequence) && dataChunkSequence.length > 0 && this.data.length > 0) {
                throw {
                    msg: 'Connector is busy.',
                    code: 'ERR_BUSY'
                }
            }
            this.status = ConnectorStatus.sending;
            if(this.data.send.length === 0) {
                await this.#SetDataChunkSequence(dataChunkSequence).then(async sequence => {
                    const startData = new ConnectionData({
                        method: 'chunk',
                        data: new ChunkStartBody({
                            type: type,
                            length: sequence.length,
                            timestamp: moment.now(),
                        })
                    });
                    const connection = await this.#GetAliveConnection();
                    await connection.Send(startData).then(res => {
                        this.#Emit('chunk_start', startData.data);
                    });
                });
            }

            interval = setInterval(() => {
                if(this.data.send.length === 0) {
                    this.thread = 0;
                    this.status = ConnectorStatus.ready;
                    clearInterval(interval);
                    return;
                }
                if(this.data.failed.length > 0) this.thread -= this.data.failed.length;
                if(this.thread >= this.threadCount) return;
                for (const key in this.connections) {
                    if(this.thread < this.threadCount && this.connections[key].status === ConnectionStatus.connected && this.connections[key].isSending === false) {
                        this.#SendData(this.connections[key], type);
                    }
                }
            }, 1000);
        } catch (e) {
            // this.#GetAliveConnectionCount(true).then(count => {
            //     if(count < this.threadCount) {
            //         this.#InitConnection({initiator: true}, false);
            //     }
            // });
            if(this.openConnectionCount < this.threadCount) this.#InitConnection({initiator: true}, false);
            if(e.code === 'ERR_SENDING' || this.data.failed.length > 0) {
                await Util.Sleep(1000);
                if(this.data.send.length > 0) this.SendDataChunkSequence();
            } else {
                this.#Emit('error', {e});
            }
        }
    }

    /**
     * disconnect all connections
     * @returns {Promise<void>}
     * @constructor
     */
    async Disconnect() {
        try {
            if(this.status === ConnectorStatus.closing || this.status === ConnectorStatus.disconnected) return;
            this.status = ConnectorStatus.closing;
            if(this.#pingingPeer) {
                clearInterval(this.#pingingPeer);
                this.#pingingPeer = null;
            }
            const connection = await this.#GetAliveConnection();
            const connectionKeys = Object.keys(this.connections);

            if(connection) {
                await connection.Send(new ConnectionData({
                    method: 'disconnect',
                    data: new DisconnectBody({connectionIds: Object.keys(this.connections)}),
                }));
                await Util.Sleep(50);
            }

            for (let i = 0; i < connectionKeys.length; i++) {
                const item = this.connections[connectionKeys[i]];
                try {
                    if(item.channel && item?.channel?.readyState === 'open' && item.status === ConnectionStatus.connected) {
                        item.Destroy();
                    }
                    this.connections[connectionKeys[i]] = null;
                } catch (e) {
                    continue;
                }
            }

            this.status = ConnectorStatus.disconnected;
        } catch (e) {
            this.#Emit('error', {e});
        }
    }

    /**
     * reset connections
     * @returns {Promise<unknown>}
     * @constructor
     */
    async Reset() {
        return new Promise(async (resolve, reject) => {
            try {
                this.openConnectionCount = 0;
                this.connections = {};
                this.On('genesis_offer', (event) => {
                    this.status = ConnectorStatus.initiated;
                    resolve(event);
                }, {once: true});
                return await this.#InitConnection({initiator: this.initiator}, this.initiator);
            } catch (e) {
                this.#Emit('error', {e});
            }
        });
    }

    /**
     * event listener -> callback
     * @param event
     * @param cb
     * @param options
     * @constructor
     */
    On(event, cb, options) {
        try {
            if(!ConnectionEvent.allowed.includes(event)) throw `Invalid event: ${event}`;
            if(Util.IsBrowser()) this.eventEmitter.addEventListener(event, cb, options);
            else this.eventEmitter.addEventListener(event, cb, options);
        } catch (e) {
            throw e;
        }
    }

    /**
     * remove event listener -> function
     * @param event
     * @param fn
     * @param options
     * @constructor
     */
    RemoveEventListener(event, fn, options) {
        try {
            if(!ConnectionEvent.allowed.includes(event)) throw `Invalid event: ${event}`;
            if(Util.IsBrowser()) this.eventEmitter.removeEventListener(event, fn, options);
            else this.eventEmitter.removeEventListener(event, fn, options);
        } catch (e) {
            throw e;
        }
    }

    /**
     * collect genesis connection answer from signal server
     * @param waitSeconds
     * @constructor
     */
    #CollectGenesisAnswer(waitSeconds = 60) {
        try {
            if(this.#collectingGenesisAnswer === true) return;

            let checkGenesisAnswerCount = 0;
            let checkGenesisAnswerErrorCount = 0;
            this.#collectingGenesisAnswer = true;
            if(this.#genesisAnswerCollectInterval) clearInterval(this.#genesisAnswerCollectInterval);
            this.#genesisAnswerCollectInterval = setInterval(() => {
                if(this.#collectingGenesisAnswer === false) {
                    clearInterval(this.#genesisAnswerCollectInterval);
                    throw {
                        code: 'ERR_CLEAR'
                    };
                }
                this.#signalserver.GetConnection(this.genesisConnectionId).then(res => {
                    if(checkGenesisAnswerCount > waitSeconds) {
                        clearInterval(this.#genesisAnswerCollectInterval);
                        this.#collectingGenesisAnswer = false;
                        throw {
                            code: 'ERR_NOANSWER'
                        };
                    }
                    if(res.answer) {
                        clearInterval(this.#genesisAnswerCollectInterval);
                        this.#collectingGenesisAnswer = false;
                        this.#Emit('genesis_answer_received', {
                            connectionId: this.genesisConnectionId,
                            answer: res.answer
                        });
                    }
                    checkGenesisAnswerCount++;
                }).catch(e => {
                    checkGenesisAnswerErrorCount++;
                    if(checkGenesisAnswerErrorCount > 10) {
                        clearInterval(this.#genesisAnswerCollectInterval);
                        this.#collectingGenesisAnswer = false;
                        throw {
                            code: 'ERR_COLLECT',
                            e
                        };
                    }
                });
            }, 1000);
        } catch (e) {
            this.#Emit('error', {e});
        }
    }

    /**
     * event emitter
     * @param name
     * @param detail
     */
    #Emit(name, detail) {
        try {
            const event = (new ConnectionEvent(name, detail)).event;
            if(Util.IsBrowser()) this.eventEmitter.dispatchEvent(event);
            else this.eventEmitter.emit(event.name, event.detail);
        } catch (e) {
            throw e;
        }
    }

    /**
     * creates a new connection and offer through an already connected connection
     * @param connection
     * @returns {Promise<*|undefined>}
     * @constructor
     */
    async #OfferConnection(connection = null) {
        try {
            if(this.status === ConnectorStatus.closing || this.status === ConnectorStatus.disconnected) return;
            if(!connection) connection = await this.#GetAliveConnection();
            if(!connection) throw {
                msg: `There is no alive connection.`,
                code: 'ERR_NOALIVECONNECTION'
            };
            return await this.#InitConnection({initiator: true}).then(async res => {
                const offer = await res.GetOffer();
                return await connection.Send(new ConnectionData({
                    method: 'offer',
                    data: new OfferBody(offer)
                }));
            });
        } catch (e) {
            if(e.code === 'ERR_NOALIVECONNECTION' && this.status !== ConnectorStatus.closing && this.status !== ConnectorStatus.disconnected) {
                this.#Emit('disconnect', {e});
            } else {
                this.#Emit('error', {e});
            }
        }
    }

    /**
     * creates a new connection
     * @param option
     * @param isGenesis
     * @returns {Promise<*>}
     * @constructor
     */
    async #InitConnection(option, isGenesis = false) {
        try {
            if(!option.wrtc) option.wrtc = this.wrtc;
            const connection = new Connection(option);
            if(isGenesis === true && (!this.connections[this.genesisConnectionId] || !this.connections[this.genesisConnectionId]?.channel)) {
                this.genesisConnectionId = connection.id;
                setTimeout(() => {
                    this.#Emit('genesis_connection', { connection });
                }, 10);
            }

            connection.OnConnect(async () => {
                this.openConnectionCount++;
                connection.Send(new ConnectionData({
                    method: 'ping',
                    data: new PingBody({
                        connectionId: connection.id,
                        peerId: connection.peerId
                    })
                }));
                if(this.openConnectionCount < this.readyConnectionCount) {
                    this.#OfferConnection(connection);
                }
            });
            connection.OnData(this.#OnDataResponse);
            connection.OnClose(async () => {
                try {
                    if(this.openConnectionCount > 0) this.openConnectionCount--;
                    if(this.openConnectionCount === 0 && this.status !== ConnectorStatus.disconnected && this.status !== ConnectorStatus.closing &&this.status !== ConnectorStatus.sending) {
                        this.#Emit('disconnect', {
                            connections: this.connections
                        });
                        return;
                    }
                    this.#OfferConnection();
                } catch (e) {
                    this.#Emit('error', {
                        msg: 'Failed to restore connection',
                        e: e,
                    });
                }
            });
            connection.OnError(async (err) => {
                try {
                    connection.Destroy(err);
                    this.connections[connection.id] = null;
                    this.#OfferConnection();
                } catch (e) {
                    e.err = err;
                    this.#Emit('error', {
                        msg: 'Failed to restore connection',
                        e: e,
                    });
                }
            });

            this.connections[connection.id] = connection;

            return this.connections[connection.id];
        } catch (e) {
            throw e;
        }
    }

    /**
     * select an alive connection
     * @returns {Promise<unknown>}
     * @constructor
     */
    async #GetAliveConnection() {
        return new Promise(async (resolve, reject) => {
            try {
                let connection = null;
                let connections = Connector._instance.connections;

                while(connection === null) {
                    for (const connectionKey in connections) {
                        if(connections[connectionKey].status === ConnectionStatus.connected && connections[connectionKey].isSending === false) {
                            connection = connections[connectionKey];
                            try {
                                await connection.Send(new ConnectionData({
                                    method: 'ping',
                                    data: new PingBody({
                                        connectionId: connection.id,
                                        peerId: connection.peerId
                                    })
                                }));
                                await Util.Sleep(10);
                                break;
                            } catch (e) {
                                connections[connectionKey].status = ConnectionStatus.error;
                                continue;
                            }
                        }
                        if(connections[connectionKey].status === ConnectionStatus.disconnected) delete connections[connectionKey];
                    }
                }

                resolve(connection);
            } catch (e) {
                resolve(null);
            }
        });
    }

    /**
     * list alive connections
     * @returns {Promise<unknown>}
     * @constructor
     */
    async #ListAliveConnection() {
        return new Promise(async (resolve, reject) => {
            try {
                let list = [];
                let connections = Connector._instance.connections;

                while(list.length === 0) {
                    for (const connectionKey in connections) {
                        if(connections[connectionKey].status === ConnectionStatus.connected && connections[connectionKey].isSending === false) {
                            let connection = connections[connectionKey];
                            try {
                                await connection.Send(new ConnectionData({
                                    method: 'ping',
                                    data: new PingBody({
                                        connectionId: connection.id,
                                        peerId: connection.peerId
                                    })
                                })).then(res => {
                                    list.push(connection);
                                });
                                await Util.Sleep(10);
                            } catch (e) {
                                connections[connectionKey].status = ConnectionStatus.error;
                                continue;
                            }
                        }
                        if(connections[connectionKey].status === ConnectionStatus.disconnected) delete connections[connectionKey];
                    }
                }

                resolve(list);
            } catch (e) {
                resolve([]);
            }
        });
    }

    /**
     * count alive connections
     * @param sendable
     * @returns {Promise<unknown>}
     * @constructor
     */
    async #GetAliveConnectionCount(sendable = false) {
        return new Promise(async (resolve, reject) => {
            try {
                let count = 0;
                let searchCount = 0;
                while(searchCount < 3 && count === 0) {
                    count = 0;
                    let connections = this.connections;
                    for (const connectionKey in connections) {
                        const aliveCondition = !connections[connectionKey].isSending || !sendable;
                        if(connections[connectionKey].status === ConnectionStatus.connected) {
                            let connection = connections[connectionKey];
                            try {
                                await connection.Send(new ConnectionData({
                                    method: 'ping',
                                    data: new PingBody({
                                        connectionId: connection.id,
                                        peerId: connection.peerId
                                    })
                                })).then(res => {
                                    if(aliveCondition) count++;
                                });
                                await Util.Sleep(10);
                            } catch (e) {
                                connections[connectionKey].status = ConnectionStatus.error;
                                continue;
                            }
                        }
                        if(connections[connectionKey].status === ConnectionStatus.disconnected) delete connections[connectionKey];
                    }
                }

                resolve(count);
            } catch (e) {
                resolve(0);
            }
        });
    }

    /**
     * data response actions
     * @param connectionId
     * @param data
     * @constructor
     */
    #OnDataResponse(connectionId, data) {
        try {
            if(typeof data === 'string') {
                try {
                    data = JSON.parse(data);
                } catch (e) {
                    throw {
                        msg: 'Invalid response.',
                        code: 'ERR_RESPONSE',
                        data
                    };
                }
            }
            if(!data.method) {
                throw {
                    msg: 'data.method not given.',
                    code: 'ERR_METHOD'
                };
            }

            switch (data.method) {
                case 'ping':
                    Connector.#OnPing(connectionId, data);
                    break;
                case 'offer':
                    Connector.#OnOffer(connectionId, data);
                    break;
                case 'answer':
                    Connector.#OnAnswer(connectionId, data);
                    break;
                case 'connect':
                    Connector.#OnConnect(connectionId, data);
                    break;
                case 'request':
                    Connector.#OnRequest(connectionId, data);
                    break;
                case 'response':
                    Connector.#OnResponse(connectionId, data);
                    break;
                case 'chunk':
                    Connector.#OnChunk(connectionId, data);
                    break;
                case 'error':
                    Connector.#OnError(connectionId, data);
                    break;
                case 'disconnect':
                    Connector.#OnDisconnect(connectionId, data);
                    break;
                default: break;
            }
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static #OnPing(connectionId, data) {
        try {
            if(data.data.timestamp) Connector._instance.lastPing = data.data.timestamp;
            if(data.data.peerId) return;
            const connection = Connector._instance.connections[connectionId];
            connection.Send(new ConnectionData({
                method: 'connect',
                data: new ConnectBody({
                    result: 'success',
                    connectionId: connectionId,
                    peerId: data.data.connectionId
                })
            }));
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static #OnError(connectionId, data) {
        try {
            Connector._instance.#Emit('error', {
                connectionId,
                data
            });
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static async #OnAnswer(connectionId, data) {
        try {
            if(!data.data.connectionId) throw `There is no connectionId`;
            const connections = Connector._instance.connections;
            const connection = connections[data.data.connectionId];
            // const connectionCount = await Connector._instance.#GetAliveConnectionCount();
            const connectionCount = Connector._instance.openConnectionCount;

            if(!connection) throw `There is no connection: ${data.data.connectionId}`;
            if(connectionCount >= Connector._instance.readyConnectionCount) return;

            await Util.Sleep(10);
            connection.SetPeerId(data.data.peerId);
            connection.SetAnswer(data.data.answer).then(res => {
                connection.Send(new ConnectionData({
                    method: 'connect',
                    data: new ConnectBody({
                        result: 'success',
                        connectionId: res.id,
                        peerId: res.peerId
                    })
                }));
            }).catch(e => {
                connection.Send(new ConnectionData({
                    method: 'error',
                    data: new ErrorBody({
                        msg: 'connect_failed',
                        error: e,
                        connectionId: connection.id,
                        peerId: data.data.connectionId
                    })
                }));
                throw e;
            });
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static async #OnOffer(connectionId, data) {
        try {
            const connections = Connector._instance.connections;
            const newConnection = await Connector._instance.#InitConnection({initiator: false});
            const connection = await Connector._instance.#GetAliveConnection();
            // const connectionCount = await Connector._instance.#GetAliveConnectionCount();
            const connectionCount = await Connector._instance.openConnectionCount;

            if(!connection) throw 'There is no working connection.';
            if(connectionCount >= Connector._instance.readyConnectionCount) return;

            await Util.Sleep(10);

            connections[newConnection.id] = newConnection;
            connections[newConnection.id].SetPeerId(data.data.connectionId);
            connections[newConnection.id].SetOffer(data.data.offer).then(answer => {
                connection.Send(new ConnectionData({
                    method: 'answer',
                    data: new AnswerBody({
                        answer,
                        connectionId: data.data.connectionId,
                        peerId: newConnection.id,
                    })
                }));
            }).catch(e => {
                connection.Send(new ConnectionData({
                    method: 'error',
                    data: new ErrorBody({
                        msg: 'offer_failed',
                        error: e,
                        connectionId: data.data.connectionId,
                        peerId: connection.id,
                    })
                }));
                throw e;
            });
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static async #OnConnect(connectionId, data) {
        try {
            if(data.data.result === 'success') {
                Connector._instance.connections[data.data.peerId].SetPeerId(data.data.connectionId);
                // const connectionCount = await Connector._instance.#GetAliveConnectionCount(true);
                const connectionCount = Connector._instance.openConnectionCount;

                if(connectionCount >= Connector._instance.readyConnectionCount) {
                    await Util.Sleep(10);
                    Connector._instance.status = ConnectorStatus.ready;
                    Connector._instance.#Emit('ready',{
                        connections: Connector._instance.connections
                    });
                }
            } else {
                throw {
                    msg: 'Connect Failed',
                    connectionId,
                    data
                }
            }
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static #OnChunk(connectionId, data) {
        try {
            Connector._instance.#Emit('chunk', data);
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static #OnRequest(connectionId, data) {
        try {
            Connector._instance.#Emit('request', data);
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static #OnResponse(connectionId, data) {
        try {
            Connector._instance.#Emit('response', data);
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    static #OnDisconnect(connectionId, data) {
        try {
            if(Connector._instance.status !== ConnectorStatus.disconnected && Connector._instance.status !== ConnectorStatus.closing) {
                Connector._instance.#Emit('disconnect', data);
            }
        } catch (e) {
            Connector._instance.#Emit('error', {e});
        }
    }

    /**
     * send loaded data
     * @param connection
     * @param type
     * @returns {Promise<unknown>}
     * @constructor
     */
    async #SendData(connection, type = 'send_tx') {
        return await new Promise(async (resolve, reject) => {
            this.thread++;
            connection.isSending = true;

            while(this.data.send.length > 0) {
                if(connection.channel !== null) {
                    const bufferedAmount = connection.channel.bufferedAmount ?? 0;
                    if(Number(bufferedAmount) > 256 * 1024) {
                        await Util.Sleep(50);
                        continue;
                    }
                }
                const order = this.data.send.length;
                const data = this.data.send.shift();
                if(!data) throw 'Invalid data';

                const sendData = new ConnectionData({
                    method: 'chunk',
                    data: new ChunkDataBody({
                        type: type,
                        data: data,
                        timestamp: moment.now(),
                        order: order
                    })
                });
                try {
                    if(connection.status === ConnectionStatus.connected) {
                        await connection.Send(sendData).then(res => {
                            this.count.sent++;
                            this.#Emit('chunk_data', sendData.data);
                        });
                    } else {
                        this.data.send = [...this.data.send, data];
                        break;
                    }
                } catch (e) {
                    connection.status = ConnectionStatus.error;
                    this.data.send.push(data);
                    this.data.failed.push(data);
                    connection.Destroy(e);
                    this.#OfferConnection();
                    break;
                }
                await Util.Sleep(50);
            }

            const endData = new ConnectionData({
                method: 'chunk',
                data: new ChunkEndBody({
                    type: type,
                    result: (this.data.send.length === 0) ? 'success' : 'failed',
                    sentCount: this.count.sent,
                    failedCount: this.data.failed.length,
                    timestamp: moment.now(),
                })
            });

            this.thread--;
            connection.isSending = false;
            if(this.status === ConnectorStatus.sending) {
                this.status = ConnectorStatus.ready;
                await connection.Send(endData).then(async res => {
                    await Util.Sleep(100);
                    this.#Emit('chunk_end', {...endData.data, failed: this.data.failed});
                });
            }

            resolve(endData.data);
        });
    }

    /**
     * load and encode array of tx objects into chunks to send
     * @param dataSequence: array of tx objects
     * @constructor
     */
    async #SetDataSequence(dataSequence) {
        return await new Promise((resolve, reject) => {
            try {
                this.data.send = [];
                for (let i = 0; i < dataSequence.length; i++) {
                    try {
                        if(typeof dataSequence[i] === 'object') dataSequence[i] = Encoder.Stringify(dataSequence[i]);
                    } catch (e) {
                        throw {
                            msg: 'Invalid item type',
                            e
                        };
                    }
                    if(dataSequence[i].length <= this.maxDataSize) {
                        this.data.send.push(new Chunk({
                            chunk: dataSequence[i],
                            order: i,
                            chunkOrder: 0,
                            chunkLength: 1,
                        }));
                        continue;
                    }
                    const chunks = fastChunkString(dataSequence[i], {size: this.maxDataSize, unicodeAware: true});
                    chunks.forEach((item, j) => {
                        this.data.send.push(new Chunk({
                            chunk: item,
                            order: i,
                            chunkOrder: j,
                            chunkLength: chunks.length,
                        }));
                    });
                }

                this.count.sent = 0;

                resolve(this.data.send);
            } catch (e) {
                reject(e);
            }
        });
    }

    /**
     * load encoded chunks of tx object array to send
     * @param dataChunkSequence: array of encoded chunks of tx object array
     * @returns {Promise<unknown>}
     * @constructor
     */
    async #SetDataChunkSequence(dataChunkSequence) {
        return await new Promise((resolve, reject) => {
            try {
                this.data.send = [];
                this.count.sent = 0;
                for (let i = 0; i < dataChunkSequence.length; i++) {
                    try {
                        if(Chunk.IsChunk(dataChunkSequence[i])) this.data.send.push(dataChunkSequence[i]);
                    } catch (e) {
                        throw {
                            msg: 'Invalid item type',
                            e
                        }
                    }
                }

                resolve(this.data.send);
            } catch (e) {
                reject(e);
            }
        });
    }

}