'use strict'

const SimplePeer = require('simple-peer');
const SaseulJS = require('saseul');
const Saseul = new SaseulJS();
const Util = require("../util/util");
const ConnectionStatus = require('./connectionstatus');
const qrcode = require("qrcode");
const Config = require("../config.js");
const ConnectionData = require("./connectiondata");
const PingBody = require("./connectiondatabody/pingbody");

module.exports = class Connection {

    #CONNECTION_ID_BYTES = 32;
    #CONNECTION_DATA_SIZE_LIMIT = 220 * 1024;
    #peer;
    channel;
    protocol = 'https';
    isSending = false;
    status = ConnectionStatus.initiated;

    id; // id of this connection
    peerId; // id of connected connection

    offer = [];
    answer = [];
    is_initiator = true;

    constructor({ initiator= true, wrtc = null }) {
        try {
            let opts = {
                initiator,
                objectMode: true,
            };
            this.id = this.#GenerateId();
            this.is_initiator = initiator;

            if(wrtc) opts.wrtc = wrtc;
            if(Util.IsBrowser()) this.protocol = location.protocol;

            this.#peer = new SimplePeer(opts);
            this.channel = this.#peer._channel;

            this.OnSignal((data) => {
                this.channel = this.#peer._channel;
                if(initiator === true) this.offer.push(data);
                else this.answer.push(data);
            });
            this.OnConnect(() => {
                this.channel = this.#peer._channel;
                this.Send(new ConnectionData({
                    method: 'ping',
                    data: new PingBody({
                        connectionId: this.id,
                        peerId: this.peerId
                    })
                }));
            });
        } catch (e) {
            throw e;
        }
    }

    /**
     * generate random hex id
     * @returns {*}
     */
    #GenerateId() {
        try {
            return Saseul.Util.Enc.RandomHexString(this.#CONNECTION_ID_BYTES);
        } catch (e) {
            throw e;
        }
    }

    /**
     * set peer connection id
     * @param id
     * @constructor
     */
    SetPeerId(id) {
        try {
            this.peerId = id ?? this.peerId;
        } catch (e) {
            throw e;
        }
    }

    /**
     * set peer connection offer
     * @param offer
     * @returns {Promise<unknown>}
     * @constructor
     */
    async SetOffer(offer) {
        try {
            return new Promise(async (resolve, reject) => {
                let count = 0;
                for (let i = 0; i < offer.length; i++) {
                    if(typeof offer[i] === 'string') offer[i] = JSON.parse(offer[i]);
                    this.Signal(offer[i]);
                    await Util.Sleep(100);
                    count++;
                }
                if(count === offer.length) resolve(this.answer);
                else reject(this);
            });
        } catch (e) {
            throw e;
        }
    }

    /**
     * set peer connection answer
     * @param answer
     * @returns {Promise<unknown>}
     * @constructor
     */
    async SetAnswer(answer) {
        try {
            return new Promise(async (resolve, reject) => {
                let count = 0;
                for (let i = 0; i < answer.length; i++) {
                    if(typeof answer[i] === 'string') answer[i] = JSON.parse(answer[i]);
                    this.Signal(answer[i]);
                    await Util.Sleep(100);
                    count++;
                }
                if(count === answer.length) resolve(this);
                else reject(this);
            });
        } catch (e) {
            throw e;
        }
    }

    /**
     * get peer connection offer
     * @returns {Promise<unknown>}
     * @constructor
     */
    async GetOffer() {
        try {
            return await new Promise((resolve, reject) => {
                const timestamp = Date.now();
                let interval = setInterval(() => {
                    if(Date.now() > timestamp + 1000) {
                        clearInterval(interval);
                        resolve({
                            connectionId: this.id,
                            offer: this.offer
                        });
                    }
                }, 100);
            });
        } catch (e) {
            throw e;
        }
    }

    /**
     * get peer connection answer
     * @returns {Promise<unknown>}
     * @constructor
     */
    async GetAnswer() {
        try {
            return await new Promise((resolve, reject) => {
                const timestamp = Date.now();
                let interval = setInterval(() => {
                    if(Date.now() > timestamp + 1000) {
                        clearInterval(interval);
                        resolve(this.answer);
                    }
                }, 100);
            });

        } catch (e) {
            throw e;
        }
    }

    /**
     * detect peer event
     * @param type
     * @param cb
     */
    #OnPeerEvent(type, cb) {
        try {
            this.#peer.on(type, cb);
        } catch (e) {
            throw e;
        }
    }

    RemovePeerEventListener(type) {
        try {
            this.#peer.removeAllListeners(type);
        } catch (e) {
            throw e;
        }
    }

    OnSignal(cb = () => {}) {
        try {
            this.#OnPeerEvent('signal', data => {
                cb(data);
            });
        } catch (e) {
            throw e;
        }
    }

    OnConnect(cb = () => {}) {
        try {
            this.#OnPeerEvent('connect', () => {
                this.status = ConnectionStatus.connected;

                cb();
            });
        } catch (e) {
            throw e;
        }
    }

    OnData(cb = () => {}) {
        try {
            this.#OnPeerEvent('data', data => {
                cb(this.id, data);
            });
        } catch (e) {
            throw e;
        }
    }

    OnClose(cb = () => {}) {
        try {
            this.#OnPeerEvent('close', () => {
                this.status = ConnectionStatus.disconnected;
                cb();
            });
        } catch (e) {
            throw e;
        }
    }

    OnError(cb = () => {}) {
        try {
            this.#OnPeerEvent('error', err => {
                this.status = ConnectionStatus.error;
                cb(err);
            });
        } catch (e) {
            throw e;
        }
    }

    Signal(data) {
        try {
            this.#peer.signal(data);
        } catch (e) {
            throw e;
        }
    }

    async Send(connectionData) {
        return await new Promise((resolve, reject) => {
            try {
                if(typeof connectionData.Object !== 'function' || typeof connectionData.Stringify !== 'function') throw `Invalid connectionData`;
                this.#peer.send(connectionData.Stringify());
                resolve();
            } catch (e) {
                this.status = ConnectionStatus.error;
                reject(e);
            }
        });
    }

    Destroy(err) {
        try {
            this.status = ConnectionStatus.disconnected;
            this.#peer.destroy(err);
        } catch (e) {
            throw e;
        }
    }

    async GetQrCodeDataURL(option = {errorCorrectionLevel: 'M', type: 'image/jpeg', margin: 1}) {
        try {
            return await qrcode.toDataURL(this.id, option);
        } catch (e) {
            throw e;
        }
    }

    GetWebURL() {
        try {
            return `${this.protocol}//${Config.GUARDEE_WEB_HOST}?connectionId=${this.id}`;
        } catch (e) {
            throw e;
        }
    }

    GetIntentURL() {
        try {
            return `${Config.GUARDEE_INTENT_HOST}?connectionId=${this.id}`;
        } catch (e) {
            throw e;
        }
    }

}